<?php

namespace LakeDrops\DockerTraefik;

use Symfony\Component\Filesystem\Filesystem;
use Symfony\Component\Yaml\Yaml;

/**
 * Class Traefik.
 *
 * @package LakeDrops\DockerTraefik
 */
class Traefik {

  /**
   * The name of the project using Traefik.
   *
   * @var string
   */
  protected $name;

  /**
   * @var string
   */
  protected $domain;

  /**
   * @var int
   */
  protected $http_port;

  /**
   * @var int
   */
  protected $https_port;

  /**
   * @var string
   */
  protected $cert_filename;

  /**
   * @var string
   */
  protected $key_filename;

  /**
   * @var bool
   */
  protected $addon_portainer = FALSE;

  /**
   * @var string
   */
  protected $hub_token = '';

  /**
   * Traefik constructor.
   *
   * @param string $name
   *   Name of the network to be created which needs to match the container
   *   prefix of your project you would like to handle with Traefik.
   * @param string $domain
   *   The domain name used for all local projects.
   * @param int $http_port
   *   Port for non secure requests.
   * @param int $https_port
   *   Port for secure requests.
   * @param string $cert_filename
   *   Filename of the SSL certificate.
   * @param string $key_filename
   *   Filename of the private key for the SSL certificate.
   */
  public function __construct(string $name, $domain = 'docker.localhost', $http_port = 8000, $https_port = 8443, $cert_filename = 'fullchain.pem', $key_filename = 'privkey.pem') {
    $this->name = $name;
    $this->domain = $domain;
    $this->http_port = $http_port;
    $this->https_port = $https_port;
    $this->cert_filename = $cert_filename;
    $this->key_filename = $key_filename;
  }

  /**
   * @param bool $addon_portainer
   */
  public function setAddonPortainer(bool $addon_portainer): void {
    $this->addon_portainer = $addon_portainer;
  }

  /**
   * @param string $hub_token
   */
  public function setHubToken(string $hub_token): void {
    $this->hub_token = $hub_token;
  }

  /**
   * Update the Traefik container.
   */
  public function update(): void {
    // Update host wide traefik container.
    $traefikPath = $_SERVER['HOME'] . '/.traefik';
    $traefikCertPath = $_SERVER['HOME'] . '/.traefik/certs';
    $traefikConfigPath = $_SERVER['HOME'] . '/.traefik/configuration';
    $traefikFile = $traefikPath . '/docker-compose.yml';

    $user = getenv('USER');
    $fs = new Filesystem();
    if (!$fs->exists($traefikPath)) {
      $fs->mkdir($traefikPath);
      $fs->chown($traefikPath, $user, TRUE);
    }
    if (!$fs->exists($traefikCertPath)) {
      $fs->mkdir($traefikCertPath);
      $fs->chown($traefikCertPath, $user, TRUE);
    }
    if (!$fs->exists($traefikConfigPath)) {
      $fs->mkdir($traefikConfigPath);
      $fs->chown($traefikConfigPath, $user, TRUE);
    }
    $upd1 = $this->updateFile($traefikConfigPath . '/certificates.toml', $this->defaultCertificatesConfig());
    $upd2 = $this->updateFile($traefikFile, Yaml::dump($this->defaultDockerCompose(), 9, 2));
    if ($upd1 || $upd2) {
      $fs->chown($traefikConfigPath . '/certificates.toml', $user, TRUE);
      $fs->chown($traefikFile, $user, TRUE);
      $cwd = getcwd();
      chdir($traefikPath);
      exec('docker network create traefik-public 2>/dev/null');
      exec('docker compose --project-name traefik up -d');
      chdir($cwd);
    }
  }

  /**
   * @param string $filename
   * @param string $content
   *
   * @return bool
   */
  private function updateFile(string $filename, string $content): bool {
    if (file_exists($filename) && file_get_contents($filename) === $content) {
      return FALSE;
    }
    file_put_contents($filename, $content);
    return TRUE;
  }

  /**
   * @return array
   */
  private function defaultDockerCompose(): array {
    $config = [
      'version' => '3.3',
      'services' => [
        'traefik' => [
          'container_name' => 'traefik',
          'image' => 'traefik:2.7',
          'command' => [
            '--api=true',
            '--api.dashboard=true',
            '--api.insecure=true',
            '--entrypoints.web.address=:80',
            '--entrypoints.websecure.address=:443',
            '--entrypoints.websecure.http.tls.domains[0].main=' . $this->domain,
            '--entrypoints.websecure.http.tls.domains[0].sans=.' . $this->domain,
            '--providers.file.directory=/configuration/',
            '--providers.file.watch=true',
            '--providers.docker=true',
            '--providers.docker.exposedbydefault=false',
          ],
          'restart' => 'unless-stopped',
          'networks' => [
            'traefik-public',
          ],
          'ports' => [
            $this->http_port . ':80',
            $this->https_port . ':443',
          ],
          'labels' => [
            'traefik.enable=true',
            'traefik.network=traefik-public',
            'traefik.http.routers.traefik.service=api@internal',
            'traefik.http.routers.traefik.rule=Host(`traefik.' . $this->domain . '`)',
          ],
          'volumes' => [
            './certs:/certs/:ro',
            './configuration:/configuration/:ro',
            '/var/run/docker.sock:/var/run/docker.sock:ro',
          ],
        ],
      ],
      'networks' => [
        'traefik-public' => [
          'external' => true,
        ],
      ],
    ];
    if ($this->addon_portainer) {
      $config['services']['portainer'] = [
        'container_name' => 'portainer',
        'image' => 'portainer/portainer-ce:2.11.1',
        'command' => '-H unix:///var/run/docker.sock',
        'restart' => 'unless-stopped',
        'networks' => [
          'traefik-public',
        ],
        'labels' => [
          'traefik.enable=true',
          'traefik.network=traefik-public',
          'traefik.http.routers.frontend.rule=Host(`portainer.' . $this->domain . '`)',
          'traefik.http.routers.frontend.entrypoints=websecure',
          'traefik.http.services.frontend.loadbalancer.server.port=9000',
          'traefik.http.routers.frontend.service=frontend',
          'traefik.http.routers.edge.rule=Host(`edge.' . $this->domain . '`)',
          'traefik.http.routers.edge.entrypoints=websecure',
          'traefik.http.services.edge.loadbalancer.server.port=8000',
          'traefik.http.routers.edge.service=edge',
        ],
        'volumes' => [
          './portainerdata:/data',
          '/var/run/docker.sock:/var/run/docker.sock:ro',
        ],
      ];
    }
    if ($this->hub_token) {
      $config['services']['traefik']['command'][] = '--experimental.hub=true';
      $config['services']['traefik']['command'][] = '--hub.tls.insecure=true';
      $config['services']['traefik']['command'][] = '--metrics.prometheus.addrouterslabels=true';
      $config['services']['traefik']['ports'][] = '9900:9900';
      $config['services']['traefik']['ports'][] = '9901:9901';
      $config['services']['hub-agent'] = [
        'container_name' => 'hub-agent',
        'image' => 'ghcr.io/traefik/hub-agent-traefik:v0.7.0',
        'command' => [
          'run',
          '--hub.token=' . $this->hub_token,
          '--auth-server.advertise-url=http://hub-agent',
          '--traefik.host=traefik',
          '--traefik.tls.insecure=true',
        ],
        'restart' => 'on-failure',
        'networks' => [
          'traefik-public',
        ],
        'volumes' => [
          '/var/run/docker.sock:/var/run/docker.sock:ro',
        ],
        'depends_on' => [
          'traefik',
        ],
      ];
    }
    return $config;
  }

  /**
   * @return string
   */
  private function defaultCertificatesConfig(): string {
    return <<<EOF
[[tls.certificates]]
   certFile = "/certs/$this->cert_filename"
   keyFile = "/certs/$this->key_filename"
EOF;

  }

}
